/*
 * xmodem.c : xmodem protrol for CP/M-80
 *  ver 0.01b at 2018/06/19 by skyriver
 *  ver 0.01c at 2018/06/24 added PUN/RDR by skyriver
 *    copyright(c) skyriver
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <conio.h>
#include <hitech.h>
#include <cpm.h>

#define	SEC_SIZE	128	/* sector size */
#define CRC_BASE	1021h	/* CRC BASE parameter */

/* bios */
char bios(int n, int a1, int a2);
#define BIOS_CONST	2
#define BIOS_CONIN	3
#define BIOS_CONOUT	4
#define	BIOS_PUNCH	6
#define	BIOS_READER	7

#define PutChar(c)	bios(BIOS_CONOUT,c,0)
#define GetChar()	bios(BIOS_CONIN,0,0)
#define KbHit()		bios(BIOS_CONST,0,0)
#define Punch(c)	bios(BIOS_PUNCH,c,0)
#define Reader()	bios(BIOS_READER,0,0)


#define C_SOH	1	/* Start of XMODEM block */
#define C_CTRLC	3	/* Control-C for user-abort */
#define C_EOT	4	/* End XMODEM session */
#define C_ACK	6	/* XMODEM block acknowledge */
#define S_BS	"\x08"	/* back space */
#define C_NAK	0x15/* XMODEM block negative ACK */
#define C_EOF	0x1a/* EOF code */
#define C_CRC	'C'	/* select crc mode */

#define TRUE	1
#define	FALSE	0

#define MAX_RETRY	12
#define RES_TIMER	20000	/* responce timer */

typedef struct {
	BYTE low;
	BYTE high;
} U16;

typedef union {
	U16 byte;
	UINT_16 word;
} UINT_BYTE;

/* groval work area */
char *FileName;
char RcvFlg;
char CrcFlg;
char PunFlg;
char RdrData;
BYTE RdrSts;
BYTE Buf[ SEC_SIZE ];
UINT_BYTE CrcVal;
UINT_BYTE BlkCnt;
FILE *Fp;
BYTE Res,SndC,Sum;
char *ErrTop = "? ";
char *CompStr[] = {
	 S_BS S_BS S_BS S_BS S_BS "t"
	,S_BS S_BS S_BS S_BS "ed"
};

char *Help = "usage : xmodem [/s /r /p] FileName\n"
    "\t/s : send[default]\n"
	"\t/r : receive\n"
	"\t/p : use PUN/RDR device\n"
	"\t    Ver0.01c 2018/06/24 by skyriver"
;


char to_upper( char c )
{
	if( (c >= 'a') && (c <= 'z') ) {
		c &= 0x5f;
	}
	return( c );
}


/*
 * calc CRC
 *  SndC <- send data
 *  CrcVal <- previos CRC data
 *  CrcVal -> updated CRC data
 */
void CalCrc( void )
{
#if 0
	int i;
	CrcVal.byte.high ^= SndC;
	for( i=8; i; i-- ) {
		if( CrcVal.word & 0x8000 ) {
			CrcVal.word <<= 1;
			CrcVal.word ^= CRC_BASE;
		} else {
			CrcVal.word <<= 1;
		}
	}
#else
#asm	/* to speed up */
;	call	.ncsv
	ld	hl,(_CrcVal)
	ld	a,(_SndC)		; get data value
	xor	h
	ld	h,a
	ld	de,CRC_BASE		; CRC_BASE
	ld	b,8
1:
	add	hl,hl
	jr	nc,2f

	ld	a,h
	xor	d
	ld	h,a
	ld	a,l
	xor	e
	ld	l,a
2:
	djnz	1b
	ld	(_CrcVal),hl
;	jp	.cret
#endasm
#endif
}


void ErrExit( char *pnt )
{
	fprintf( stderr,"%s %s", ErrTop, pnt );
	if( Fp != NULL ) {
		fclose( Fp );
	}
	exit( 1 );
}


/*
 * send character
 */
void SendChar( BYTE c )
{
	if( PunFlg == FALSE ) {
		PutChar( c );
	} else {
		Punch( c );
	}
}


/*
 * send block data
 *  c <- data
 */
void SendData( BYTE c )
{
	if( CrcFlg ) {
		SndC = c;
		CalCrc();
	} else {
		Sum += c;
	}
	SendChar( c );
}


/*
 * get Reader status
 * return -> 0:no data
 */
BYTE RdrStatus( void )
{
	Reader();
#asm
		jr	z,1f
		ld	(_RdrData),a
		ld	a,1
1:	ld	(_RdrSts),a
#endasm
	return( RdrSts );
}


BYTE RcvChar( void )
{
	BYTE retval;
	if( PunFlg == FALSE ) {
		retval = GetChar();
	} else {
		while( RdrStatus() == 0 );
		retval = RdrData;
	}
	return( retval );
}

/*
 * get responce
 *  res <- responce pointer
 *  *res -> responce
 *  return -> 0:time over
 */
UINT_16 GetRes( BYTE *res )
{
	UINT_16 tim;
	int i;

	*res = 0xff;
	tim = RES_TIMER;
	while( (--tim != 0) && ( *res == 0xff ) ) {
		for( i= PunFlg == FALSE ? 4 : 3 ; i; i-- ) {
			if( PunFlg == FALSE ) {
				if( KbHit() != 0 ) {
					*res = GetChar();
					i = 1;
				}
			} else {
				if( RdrStatus() != 0 ) {
					*res = RdrData;
					i = 1;
				}
			}
		}
	}
	return( tim );
}


/*
 * send file with xmodem
 */
int Send( void )
{
	int i,rdlen;
	BYTE *pnt, retry;

	if( ( Fp = fopen( FileName, "rb" )) == NULL ) {
		ErrExit( "cann't open" );
	}
	retry = MAX_RETRY;
	Res = 0;
	fprintf( stderr, "sending " );
	while( ( retry != 0 ) && (Res != C_NAK) && (Res != C_CRC) ) {
		retry--;
		GetRes( &Res );
	}
	if( Res == C_CRC ) {
		CrcFlg = TRUE;
	}
	else if( Res != C_NAK ) {
		ErrExit( "bad responce" );
	}
	Res = 0;
	do {
		if( Res != C_NAK ) {
			rdlen = fread( Buf, sizeof(Buf[0]), SEC_SIZE, Fp );
		}
		if( rdlen > 0 ) {	/* not EOF */
			BlkCnt.word++;
			SendChar( C_SOH );
			SendChar( BlkCnt.byte.low );
			SendChar( ~BlkCnt.byte.low );
			CrcVal.word = 0;
			Sum = 0;
			for( i=0, pnt = Buf; i< rdlen; i++ ) {
				SendData( *pnt++ );
			}
			for( ; i< SEC_SIZE; i++ ) {
				SendData( C_EOF );		/* if CP/M then not execute this parts */
			}
			if( CrcFlg != FALSE ) {
				SendChar( CrcVal.byte.high );
				SendChar( CrcVal.byte.low );
			} else {
				SendChar( Sum );
			}

		} else {	/* EOF */
			SendChar( C_EOT );
		}
		if( GetRes( &Res ) == 0 ) {
			ErrExit("no Res" );
		}
		else if( Res == C_NAK ) {
			BlkCnt.word--;
		}
		else if( Res != C_ACK ) {
			ErrExit( "  irregular Res" );
		}
	} while( rdlen );
	fclose( Fp );
	return( 0 );
}


/*
 * receive file with xmodem
 */
int Receive( void )
{
	int i;
	BYTE c, *pnt, type, errcnt, retry;
	if( ( Fp = fopen( FileName, "r" ) ) != NULL ) {
		fclose( Fp );
		fprintf( stderr, " Overwrite(Y/N) ?" );
		c = bios( BIOS_CONIN,0,0 );
		PutChar( c );
		if( to_upper( c ) != 'Y' ) {
			ErrExit( " bye!" );
		}
		unlink( FileName );
	}
	if( ( Fp = fopen( FileName, "wb" ) ) == NULL ) {
		ErrExit( "cann't open" );
	}
	BlkCnt.word = 0;
	Res = C_NAK;
	fprintf( stderr, "  receiving " );

	do {
		retry = MAX_RETRY;
		type = 0xff;
		do {
			SendChar( Res );
			if( GetRes( &type ) == 0 ) {
				retry--;
			}
		} while( ( retry != 0 ) && type == 0xff );
		if( retry == 0 ) {
				ErrExit( "time over" );
		}
		if( type == C_SOH ) {
			errcnt = 0;
			BlkCnt.word++;
			if( RcvChar() != BlkCnt.byte.low ) {
				errcnt++;
			}
			if( RcvChar() != ~BlkCnt.byte.low ) {
				errcnt++;
			}
			Sum = 0;
			for( i = SEC_SIZE, pnt = Buf; i; i-- ) {
				c = RcvChar();
				Sum += c;
				*pnt++ = c;
			}
			if( Sum !=  RcvChar() ) {
				errcnt++;
			}
			if( errcnt == 0 ) {
				if( fwrite( Buf, sizeof(Buf[0]), SEC_SIZE, Fp ) == SEC_SIZE ) {
					Res = C_ACK;
				}
			} else {
				Res = C_NAK;
				BlkCnt.word--;
			}
		}
		else if( type == C_EOT ) {
			SendChar( C_ACK );
		} else {
			ErrExit( "packet err" );
		}
	} while( type != C_EOT );
	fclose( Fp );
	return( 0 );
}


int main( int argc, char *argv[] )
{
	char *pnt;
	int retval;

	if( argc == 1 ) {	/* no arg */
		fprintf( stderr, Help );
		exit( 0 );
	}
	while( --argc ) {
		pnt = *++argv;
		if( *pnt == '/' ) {
			switch( to_upper(pnt[1]) ) {
			case 'S':	/* send */
				RcvFlg = FALSE;
				break;
			case 'R':	/* receive */
				RcvFlg = TRUE;
				break;
			case 'P':	/* PUN:/RDR: */
				PunFlg = TRUE;
				break;
			default:
				ErrExit( pnt );
				break;
			}
		} else {
			if( FileName == NULL ) {
				FileName = pnt;
			} else {
				ErrExit( pnt );
			}
		}
	}
	if( FileName == NULL ) {
		ErrExit( "set file name" );
	}
	*ErrTop = '\0';
	fprintf( stderr, "%s ", FileName );
	if( PunFlg != FALSE ) {
		fprintf( stderr," (PUN/RDR) " );
	}
	if( RcvFlg ) {
		retval = Receive();
	} else {
		retval = Send();
	}
	if( retval == 0 ) {	/* no error */
		fprintf( stderr, CompStr[ RcvFlg ] );
		fprintf( stderr, " %u block(s).", BlkCnt.word );
	}
	return( retval );
}
